#include "CubeViewUI.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#define PI2 1.570796327
#define maxnd 3
#define COULOMB 1.0
#define INVMASS 1.0

#define POTEXP 8
#define POTINVRAD 0.25

/***** the potential *****/
double v(double x) {

#ifdef COULOMB  

  return COULOMB/x;

#else

  double h;

  h = x<PI2 ? x : PI2;
 
  h = sin(h);

  return h*h;

#endif

}

/***** derivative of the potential *****/
double dv(double x) {

#ifdef COULOMB

  return -COULOMB/(x*x);

#else

  double h;
  
  h = x<PI2 ? x : PI2;

  return 2.0*sin(h)*cos(h);

#endif

}

/***** initialize *****/
void initialize(int np, int nd, double *box, double *pos, double *vel) {

  int i, j;
  
  for(i=0; i<np; i++)
    for(j=0; j<nd; j++) {
      pos[nd*i + j] = box[j]*(2.0*drand48() - 1.0);
      vel[nd*i + j] = 0.0;
    }
 
}

/***** calculate distance and directions *****/
double dist(int nd, double *r1, double *r2, double *dr) {

  int j;
  double d;
  
  d = 0.0;
  for(j=0; j<nd; j++) { 
    dr[j] = r1[j] - r2[j];
    d += dr[j]*dr[j];
  }
  
  return sqrt(d);

}

/***** calculate forces *****/
void force(int np, int nd, double *pos, double *f, double *pot) {

  int i, j, k, ni;
  double d, ex, pn, lpot, rij[maxnd];
  
  lpot = 0.0;
  ni = np*nd;
  
#pragma omp parallel for default(shared) private(i,j,k,d,ex,rij) reduction(+:lpot)
  for(i=0; i<ni; i+=nd) {
    
    // external forces (walls)
    for(j=0; j<nd; j++) {
      d = POTINVRAD*pos[i+j];
      
      ex = 1.0;
      for(k=0; k<POTEXP-1; k++) ex *= d;
      
      lpot += ex*d;
      
      f[i+j] = -POTINVRAD*((double)POTEXP)*ex;
    }

    // internal forces
    for(j=0; j<ni; j+=nd) {
      if(i!=j) {
	d = dist(nd, &pos[i], &pos[j], rij);
	lpot += 0.5*v(d);
	for(k=0; k<nd; k++) f[i+k] -= rij[k]*dv(d)/d;
      }
    } 
  }

  *pot = lpot;
    

}

/***** position and velocity updates *****/
void update(int np, int nd, double *pos, double *vel, double *f, double dt, double *pot, double *kin) {

  int i, j, ni;
  double d, dth, lkin;

  lkin = 0.0;
  dth = 0.5*dt*INVMASS;
  ni = np*nd;

#pragma omp parallel for default(shared) private(i)
  for(i=0; i<ni; i++) {
    vel[i] += dth*f[i];
    pos[i] += dt*vel[i];
  }
  
  force(np, nd, pos, f, pot);

#pragma omp parallel for default(shared) private(i,d) reduction(+:lkin)
  for(i=0; i<ni; i++) {
    d = vel[i] + dth*f[i];
    vel[i] = d;
    lkin += d*d;
  }  

  *kin = 0.5*lkin/INVMASS;

}

/***** error handling *****/
void error_abort(char *s) {
  
  printf("Error: %s\n", s);
  exit(0);

}

/***** main driver *****/
int main(int argc, char **argv) {

  int ndim = 3;
  int nparts, nsteps, i, j;

  double mass, dt, potential, kinetic;
  double *box, *position, *velocity, *force;

  srand48(time(NULL));

  dt = 1.0e-3;
  mass = 1.0;
  nparts = 100;
  nsteps = 5;
  
  if((box = (double *)calloc(ndim, sizeof(double))) == NULL) 
    error_abort("Could not allocate memory for box.");
  if((position = (double *)calloc(ndim*nparts, sizeof(double))) == NULL) 
    error_abort("Could not allocate memory for position.");
  if((velocity = (double *)calloc(ndim*nparts, sizeof(double))) == NULL) 
    error_abort("Could not allocate memory for velocity.");
  if((force = (double *)calloc(ndim*nparts, sizeof(double))) == NULL) 
    error_abort("Could not allocate memory for force.");
 
  /* start up */
  for(i=0; i<ndim; i++) box[i] = 2.0;
  initialize(nparts, ndim, box, position, velocity);
  
  CubeViewUI *mycube = new CubeViewUI();

  (mycube->cube)->buckysize = 0.1;
  (mycube->cube)->buckynum = nparts;
  (mycube->cube)->buckyposition = position;

  mycube->show(argc,argv);

  while(Fl::check()) {

    for(i=0; i<nsteps; i++) update(nparts,ndim,position,velocity,force,dt,&potential,&kinetic);

    // printf(" %lg %lg\n", potential, kinetic);
    
    (mycube->cube)->redraw();
    
  }

  free(box);
  free(position);
  free(velocity);
  free(force);

  return 0;

}
